import {BaseSchema} from 'valibot';
import {
	ApplicationState,
	BaseResourceType,
	ContentNegotiation,
	Language,
	LanguageStatusMessageMap,
	Resource,
} from '../common';
import {DataSource} from './data-source';

export interface BackendState {
	app: ApplicationState;
	dataSource: DataSource;
	cn: ContentNegotiation;
	showTotalItemCountOnGetCollection: boolean;
	throwsErrorOnDeletingNotFound: boolean;
	checksSerializersOnDelete: boolean;
	showTotalItemCountOnCreateItem: boolean;
}

export interface RequestContext {}

export interface Middleware {}

export class MiddlewareError extends Error {}

export interface MiddlewareResponseErrorParams extends Omit<Response, 'statusMessage'> {
	cause?: unknown;
}

export abstract class MiddlewareResponseError extends MiddlewareError implements Response {
	readonly statusMessage: Response['statusMessage'];
	readonly statusCode: Response['statusCode'];
	readonly headers: Response['headers'];

	constructor(statusMessage: keyof Language['statusMessages'], params: MiddlewareResponseErrorParams) {
		super(statusMessage, { cause: params.cause });
		this.statusCode = params.statusCode;
		this.headers = params.headers;
		this.statusMessage = statusMessage;
	}
}

export type RequestDecorator = (req: RequestContext) => RequestContext | Promise<RequestContext>;

export type ParamRequestDecorator<Params extends Array<unknown> = []> = (...args: Params) => RequestDecorator;

// TODO put this in HTTP
export type Method = 'GET' | 'POST' | 'PUT' | 'PATCH' | 'DELETE' | 'OPTIONS' | 'HEAD' | 'QUERY';

export interface AllowedMiddlewareSpecification<Schema extends BaseSchema = BaseSchema> {
	method: Method;
	middleware: Middleware;
	constructBodySchema?: (resource: Resource<BaseResourceType & { schema: Schema }>, resourceId?: string) => BaseSchema;
	allowed: (resource: Resource<BaseResourceType & { schema: Schema }>) => boolean;
}

export interface Response {
	// type of response
	statusCode: number;

	// description of response
	statusMessage?: keyof LanguageStatusMessageMap;

	// metadata of the response
	headers?: Record<string, string>;
}

export const getAllowString = (middlewares: AllowedMiddlewareSpecification[]) => {
	const allowedMethods = middlewares.flatMap((m) => m.method === 'GET' ? [m.method, 'HEAD'] : [m.method]);
	return allowedMethods.join(',');
}
