import {
	QueryMediaType,
	QueryAndGrouping,
	QueryAnyExpression,
	QueryOrGrouping,
} from '../../common';

import {
	DeserializeError,
	SerializeError,
} from '../../errors';

interface ProcessEntryBase {
	type: string;
}

const AVAILABLE_MATCH_TYPES = [
	'startsWith',
	'endsWith',
	'includes',
	'regexp',
] as const;

type MatchType = typeof AVAILABLE_MATCH_TYPES[number];

interface ProcessEntryString extends ProcessEntryBase {
	type: 'string';
	matchType?: MatchType;
	caseInsensitiveMatch?: boolean;
}

interface ProcessEntryNumber extends ProcessEntryBase {
	type: 'number';
	decimal?: boolean;
}

interface ProcessEntryBoolean extends ProcessEntryBase {
	type: 'boolean';
	truthyStrings?: string[];
}

type ProcessEntry = ProcessEntryString | ProcessEntryNumber | ProcessEntryBoolean;

export const name = 'application/x-www-form-urlencoded' as const;

class DeserializeInvalidFormatError extends DeserializeError {}

const normalizeRhs = (lhs: string, rhs: string, processEntriesMap?: Record<string, ProcessEntry>) => {
	const defaultCoerceValues = {
		type: 'string'
	} as ProcessEntry;
	const coerceValues = processEntriesMap?.[lhs] ?? defaultCoerceValues;

	if (coerceValues?.type === 'number') {
		return {
			lhs,
			operator: '=',
			rhs: Number(rhs)
		} as QueryAnyExpression;
	}

	if (coerceValues?.type === 'boolean') {
		const truthyStrings = [
			...(coerceValues.truthyStrings ?? []).map((s) => s.trim().toLowerCase()),
			'true',
		];

		return {
			lhs,
			operator: '=',
			rhs: truthyStrings.includes(rhs)
		} as QueryAnyExpression;
	}

	if (coerceValues?.type === 'string') {
		switch (coerceValues?.matchType) {
			case 'startsWith': {
				return {
					lhs,
					operator: coerceValues.caseInsensitiveMatch ? 'ILIKE' : 'LIKE',
					rhs: `%${rhs}`,
				} as QueryAnyExpression;
			}
			case 'endsWith': {
				return {
					lhs,
					operator: coerceValues.caseInsensitiveMatch ? 'ILIKE' : 'LIKE',
					rhs: `${rhs}%`,
				} as QueryAnyExpression;
			}
			case 'includes': {
				return {
					lhs,
					operator: coerceValues.caseInsensitiveMatch ? 'ILIKE' : 'LIKE',
					rhs: `%${rhs}%`,
				} as QueryAnyExpression;
			}
			case 'regexp': {
				return {
					lhs,
					operator: 'REGEXP',
					rhs: new RegExp(rhs, coerceValues.caseInsensitiveMatch ? 'i' : ''),
				} as QueryAnyExpression;
			}
			default:
				break;
		}

		return {
			lhs,
			operator: '=',
			rhs,
		} as QueryAnyExpression;
	}

	const unknownCoerceValues = coerceValues as unknown as Record<string, string>;
	throw new DeserializeInvalidFormatError(`Invalid coercion type: ${unknownCoerceValues.type}`);
	// this will be sent to the data source, e.g., the SQL query
	// we can also make this function act as a "sanitizer"
}

interface SerializeOptions {}

interface DeserializeOptions {
	processEntries?: Record<string, ProcessEntry>;
}

const doesGroupHaveExpression = (ex2: QueryAnyExpression, key: string) => {
	if ('operator' in ex2) {
		return ex2.lhs === key;
	}

	if ('name' in ex2) {
		return ex2.name === key;
	}

	return false;
};

export const deserialize: QueryMediaType<
	typeof name,
	SerializeOptions,
	DeserializeOptions
>['deserialize'] = (s: string, options = {} as DeserializeOptions) => {
	const q = new URLSearchParams(s);

	return Array.from(q.entries()).reduce(
		(queries, [key, value]) => {
			const defaultOr = {
				type: 'or',
				expressions: [],
			} as QueryOrGrouping;
			const existingOr = queries.expressions.find((ex) => (
				ex.expressions.some((ex2) => doesGroupHaveExpression(ex2, key))
			)) ?? defaultOr;
			const existingLhs = existingOr.expressions.find((ex) => doesGroupHaveExpression(ex, key));
			const newExpression = normalizeRhs(key, value, options.processEntries);

			if (typeof existingLhs === 'undefined') {
				return {
					...queries,
					expressions: [
						...queries.expressions,
						{
							type: 'or',
							expressions: [
								newExpression,
							],
						},
					],
				};
			}

			return {
				...queries,
				expressions: queries.expressions.map((ex) => (
					ex.expressions.some((ex2) => !doesGroupHaveExpression(ex2, key))
						? ex
						: {
							...existingOr,
							expressions: [
								...(existingOr.expressions ?? []),
								newExpression,
							],
						}
				)),
			};
		},
		{
			type: 'and',
			expressions: [],
		} as QueryAndGrouping
	)
};

class SerializeInvalidExpressionError extends SerializeError {}

const serializeExpression = (ex2: QueryAnyExpression) => {
	if ('name' in ex2) {
		return [ex2.name, `(${ex2.args.map((s) => s.toString()).join(',')})`];
	}

	if (ex2.rhs instanceof RegExp) {
		if (ex2.operator !== 'REGEXP') {
			throw new SerializeInvalidExpressionError(`Invalid rhs given for operator: ${ex2.lhs} ${ex2.operator} <rhs>`);
		}

		return [ex2.lhs, ex2.rhs.toString()];
	}

	switch (typeof ex2.rhs) {
		case 'string': {
			switch (ex2.operator) {
				case 'ILIKE':
				case 'LIKE':
					return [ex2.lhs, ex2.rhs.replace(/^%+/, '').replace(/%+$/, '')];
				case '=':
					return [ex2.lhs, ex2.rhs];
				default:
					break;
			}
			throw new SerializeInvalidExpressionError(`Invalid operator given for lhs: ${ex2.lhs} <op> ${ex2.rhs}`);
		}
		case 'number': {
			return [ex2.lhs, ex2.rhs.toString()];
		}
		case 'boolean': {
			return [ex2.lhs, ex2.rhs ? 'true' : 'false'];
		}
		default:
			break;
	}

	throw new SerializeInvalidExpressionError(`Unknown type for rhs: ${ex2.lhs} ${ex2.operator} <rhs>`);
};

export const serialize: QueryMediaType<
	typeof name,
	SerializeOptions,
	DeserializeOptions
>['serialize'] = (q: QueryAndGrouping) => (
	new URLSearchParams(
		q.expressions.flatMap((ex) => (
			ex.expressions.map((ex2) => serializeExpression(ex2))
		))
	)
		.toString()
);
