import { readFile, writeFile } from 'fs/promises';
import { join } from 'path';
import { Resource, validation as v, BaseResourceType } from '@modal-sh/yasumi';
import { DataSource, ResourceIdConfig } from '@modal-sh/yasumi/backend';
import assert from 'assert';

export class ResourceNotPreparedError extends Error {}

export class ResourceIdNotDesignatedError extends Error {}

export class JsonLinesDataSource<
  Schema extends v.BaseSchema = v.BaseSchema,
  CurrentName extends string = string,
  CurrentRouteName extends string = string,
  Data extends object = v.Output<Schema>,
> implements DataSource<Data> {
  private path?: string;

  private resource?: Resource<BaseResourceType & {
    schema: Schema,
    name: CurrentName,
    routeName: CurrentRouteName,
  }>;

  data: Data[] = [];

  constructor(private readonly baseDir = '') {
    // noop
  }

  prepareResource<Schema extends v.BaseSchema>(resource: Resource<BaseResourceType & {
    schema: Schema,
    name: CurrentName,
    routeName: CurrentRouteName,
  }>) {
    this.path = join(this.baseDir, `${resource.state.routeName}.jsonl`);
    resource.dataSource = resource.dataSource ?? this;
    const originalResourceId = resource.id;
    resource.id = <NewIdAttr extends BaseResourceType['idAttr'], NewIdSchema extends BaseResourceType['idSchema']>(newIdAttr: NewIdAttr, params: ResourceIdConfig<NewIdSchema>) => {
      originalResourceId(newIdAttr, params);
      return resource as Resource<BaseResourceType & {
        name: CurrentName,
        routeName: CurrentRouteName,
        schema: Schema,
        idAttr: NewIdAttr,
        idSchema: NewIdSchema,
      }>;
    };
    this.resource = resource as any;
  }

  async initialize() {
    assert(typeof this.path === 'string', new ResourceNotPreparedError());

    try {
      const fileContents = await readFile(this.path, 'utf-8');
      const lines = fileContents.split('\n');
      this.data = lines.filter((l) => l.trim().length > 0).map((l) => JSON.parse(l));
    } catch (err) {
      await writeFile(this.path, '');
    }
  }

  async getTotalCount() {
    return this.data.length;
  }

  async getMultiple() {
    return [...this.data];
  }

  async newId() {
    const idConfig = this.resource?.state.shared.get('idConfig') as ResourceIdConfig<any>;
    assert(typeof idConfig !== 'undefined', new ResourceNotPreparedError());
    const theNewId = await idConfig.generationStrategy(this);
    return theNewId as string;
  }

  async getById(idSerialized: string) {
    assert(typeof this.resource !== 'undefined', new ResourceNotPreparedError());
    assert(typeof this.path === 'string', new ResourceNotPreparedError());

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string', new ResourceIdNotDesignatedError());

    const idConfig = this.resource.state.shared.get('idConfig') as ResourceIdConfig<any> | undefined;
    assert(typeof idConfig !== 'undefined', new ResourceIdNotDesignatedError());

    const id = idConfig.deserialize(idSerialized);
    const foundData = this.data.find((s) => (s as any)[idAttr] === id);

    if (foundData) {
      return {
        ...foundData
      };
    }

    return null;
  }

  async create(data: Data) {
    assert(typeof this.resource !== 'undefined', new ResourceNotPreparedError());
    assert(typeof this.path === 'string', new ResourceNotPreparedError());

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string', new ResourceIdNotDesignatedError());

    const idConfig = this.resource.state.shared.get('idConfig') as ResourceIdConfig<any> | undefined;
    assert(typeof idConfig !== 'undefined', new ResourceIdNotDesignatedError());

    let theId: any;
    const { [idAttr]: dataId, ...etcData } = data as Record<string, unknown>;
    if (typeof dataId !== 'undefined') {
      theId = idConfig.deserialize((data as Record<string, string>)[idAttr]);
    } else {
      const newId = await this.newId();
      theId = idConfig.deserialize(newId);
    }
    const newData = {
      [idAttr]: theId,
      ...etcData
    } as Record<string, unknown>;

    const now = Date.now(); // TODO how to serialize dates
    const createdAt = this.resource.state.shared.get('createdAtAttr');
    if (typeof createdAt === 'string') {
      newData[createdAt] = now;
    }

    const updatedAt = this.resource.state.shared.get('updatedAtAttr');
    if (typeof updatedAt === 'string') {
      newData[updatedAt] = now;
    }

    const newCollection = [
      ...this.data,
      newData
    ];

    await writeFile(this.path, newCollection.map((d) => JSON.stringify(d)).join('\n'));

    return newData as Data;
  }

  async delete(idSerialized: string) {
    assert(typeof this.resource !== 'undefined', new ResourceNotPreparedError());
    assert(typeof this.path === 'string', new ResourceNotPreparedError());

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string', new ResourceIdNotDesignatedError());

    const idConfig = this.resource.state.shared.get('idConfig') as ResourceIdConfig<any> | undefined;
    assert(typeof idConfig !== 'undefined', new ResourceIdNotDesignatedError());

    const oldDataLength = this.data.length;

    const id = idConfig.deserialize(idSerialized);
    const newData = this.data.filter((s) => !((s as any)[idAttr] === id));

    await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));

    return oldDataLength !== newData.length;
  }

  async emplace(idSerialized: string, dataWithId: Data) {
    assert(typeof this.resource !== 'undefined', new ResourceNotPreparedError());
    assert(typeof this.path === 'string', new ResourceNotPreparedError());

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string', new ResourceIdNotDesignatedError());

    const idConfig = this.resource.state.shared.get('idConfig') as ResourceIdConfig<any> | undefined;
    assert(typeof idConfig !== 'undefined', new ResourceIdNotDesignatedError());

    const existing = await this.getById(idSerialized);
    const id = idConfig.deserialize(idSerialized);
    const { [idAttr]: idFromResource, ...data } = (dataWithId as any);
    const dataToEmplace = {
      [idAttr]: id,
      ...data,
    } as Record<string, unknown>;

    if (existing) {
      const createdAt = this.resource.state.shared.get('createdAtAttr');
      if (typeof createdAt === 'string') {
        dataToEmplace[createdAt] = (existing as Record<string, unknown>)[createdAt];
      }

      const now = Date.now(); // TODO how to serialize dates
      const updatedAt = this.resource.state.shared.get('updatedAtAttr');
      if (typeof updatedAt === 'string') {
        dataToEmplace[updatedAt] = now;
      }

      const newData = this.data.map((d) => {
        if ((d as any)[idAttr] === id) {
          return dataToEmplace;
        }

        return d;
      });

      await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));

      return [dataToEmplace, false] as [Data, boolean];
    }

    const newData = await this.create(dataToEmplace as Data);
    return [newData, true] as [Data, boolean];
  }

  async patch(idSerialized: string, data: Partial<Data>) {
    assert(typeof this.resource !== 'undefined', new ResourceNotPreparedError());
    assert(typeof this.path === 'string', new ResourceNotPreparedError());

    const idAttr = this.resource.state.shared.get('idAttr');
    assert(typeof idAttr === 'string', new ResourceIdNotDesignatedError());

    const idConfig = this.resource.state.shared.get('idConfig') as ResourceIdConfig<any> | undefined;
    assert(typeof idConfig !== 'undefined', new ResourceIdNotDesignatedError());

    const existing = await this.getById(idSerialized);
    if (!existing) {
      return null;
    }

    const newItem = {
      ...existing,
      ...data,
    } as Record<string, unknown>;

    const createdAt = this.resource.state.shared.get('createdAtAttr');
    if (typeof createdAt === 'string') {
      newItem[createdAt] = (existing as Record<string, unknown>)[createdAt];
    }

    const now = Date.now(); // TODO how to serialize dates
    const updatedAt = this.resource.state.shared.get('updatedAtAttr');
    if (typeof updatedAt === 'string') {
      newItem[updatedAt] = now;
    }

    const id = idConfig.deserialize(idSerialized);
    const newData = this.data.map((d) => {
      if ((d as any)[idAttr] === id) {
        return newItem;
      }

      return d;
    });

    await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));
    return newItem as Data;
  }
}
